﻿Imports System.Runtime.CompilerServices, System.Drawing
Public Module ControlBehavior

#Region "Button"

    Public Sub DisableButton(ParamArray btns As Button())
        btns.ForEach(Sub(btn) DisableButton(btn))
    End Sub

    <Extension>
    Public Sub DisableButton(btn As Button)
        btn.Enabled = False
    End Sub

    Public Sub EnableButton(ParamArray btns As Button())
        btns.ForEach(Sub(btn) EnableButton(btn))
    End Sub

    <Extension>
    Public Sub EnableButton(btn As Button)
        btn.Enabled = True
    End Sub

#End Region 'Button

#Region "CheckBox"

    Public Sub EnableCb(ParamArray cbs As CheckBox())
        cbs.ForEach(Sub(x) x.Enabled = True)
    End Sub

    Public Sub DisableCb(ParamArray cbs As CheckBox())
        cbs.ForEach(Sub(x) x.Enabled = False)
    End Sub

#End Region 'CheckBox

#Region "ComboBox"

    Public Sub EnableDisableCombobox(enable As Boolean, ParamArray cmbs As ComboBox())
        cmbs.ForEach(Sub(cmb) cmb.Enabled = enable)
    End Sub

#End Region 'ComboBox

#Region "GroupBox"

    Public Sub DisableGroupBox(ParamArray gbs As GroupBox())
        gbs.ForEach(Sub(gb) DisableGroupBox(gb))
    End Sub

    <Extension>
    Public Sub DisableGroupBox(gb As GroupBox)
        gb.Enabled = False
    End Sub

#End Region 'GroupBox

#Region "RadioButton"

    Public Sub DisableRbtn(ParamArray rbtns As RadioButton())
        rbtns.ForEach(Sub(x) x.Enabled = False)
    End Sub

    Public Sub EnableRbtn(ParamArray rbtns As RadioButton())
        rbtns.ForEach(Sub(x) x.Enabled = True)
    End Sub

#End Region 'RadioButton

#Region "TextBox"

    Public Sub EnableDisableTbReadOnly(tbReadOnly As Boolean, ParamArray tbs As TextBox())
        tbs.ForEach(Sub(tb) tb.ReadOnly = tbReadOnly)
    End Sub

#Region "Textbox Enter_Click_SelectAll"

    Public Sub RegisterTextBoxSelectAll(ParamArray textboxes As TextBox())
        For Each tb In textboxes
            'AddHandler tb.Enter, AddressOf TextBox_Enter
            AddHandler tb.MouseDoubleClick, AddressOf TextBox_MouseDoubleClick
        Next
    End Sub

#End Region 'Textbox Enter_Click_SelectAll

#Region "Pre-Text-Textbox"

    ''' <summary> Ermöglicht die Angabe von einer Art Hinweistext in einer Textbox. Dieser Hinweis verschwindet beim TextBox_Enter ''' </summary>
    Public Sub RegisterPretextTextbox(ParamArray TextBoxes As TextBox())
        For Each tb In TextBoxes
            AddHandler tb.Enter, AddressOf ptTb_Enter
            AddHandler tb.Leave, AddressOf ptTb_Leave
        Next
    End Sub

    Private _preText As String

    Private Sub ptTb_Enter(sender As Object, e As EventArgs)
        Dim tb = DirectCast(sender, TextBox)
        If tb.ForeColor = SystemColors.GrayText Then
            _preText = tb.Text
            tb.Text = ""
            tb.ForeColor = SystemColors.WindowText
        End If
    End Sub

    Private Sub ptTb_Leave(sender As Object, e As EventArgs)
        Dim tb = DirectCast(sender, TextBox)
        If tb.Text = "" Then
            tb.Text = _preText
            tb.ForeColor = SystemColors.GrayText
        End If
    End Sub

#End Region 'Pre-Text-Textbox

#Region "DecimalTextBox"

    ''' <summary> Bewirkt, dass nur Dezimalzahlen (mit Komma) in Textboxen akzeptiert werden </summary>
    Public Sub RegisterDecimalInput(ParamArray decimalTextBoxes As TextBox())
        For Each dctb In decimalTextBoxes
            AddHandler dctb.KeyDown, AddressOf dctb_KeyDown
            AddHandler dctb.Enter, AddressOf TextBox_Enter
            AddHandler dctb.MouseClick, AddressOf TextBox_MouseDoubleClick
            AddHandler dctb.Leave, AddressOf TextBox_numericLeave
        Next
    End Sub

    Private Sub dctb_KeyDown(sender As Object, e As KeyEventArgs)
        Dim tb = DirectCast(sender, TextBox)
        Select Case e.KeyCode
            Case Keys.Space : tb.Text = Date.Today.Year.ToString
            Case Keys.D0 To Keys.D9, Keys.NumPad0 To Keys.NumPad9, Keys.Decimal, Keys.Back, Keys.Subtract, Keys.OemMinus, Keys.Oemcomma 'nur Ziffern "," "-" und Backspace zulassen
            Case Keys.Shift, Keys.Home, Keys.End, Keys.Delete, Keys.Left, Keys.Right, Keys.OemMinus
            Case Else : e.SuppressKeyPress = True 'alle anderen Eingaben ablehnen
        End Select
    End Sub

#End Region 'DecimalTextBox

#Region "PlzTextBox"

    ''' <summary> Bewirkt, dass nur Nummern und "x" in Textboxen akzeptiert werden. Außerdem wird die Input-Länge auf 5 limitiert </summary>
    Public Sub RegisterPlzInput_PlzAsString(ParamArray plzTextBoxes As TextBox())
        For Each plztb In plzTextBoxes
            plztb.MaxLength = 5
            AddHandler plztb.KeyDown, AddressOf PlzTextBox_KeyDown
            AddHandler plztb.Enter, AddressOf TextBox_Enter
            AddHandler plztb.MouseClick, AddressOf TextBox_MouseDoubleClick
            AddHandler plztb.Leave, AddressOf plztb_Leave
        Next
    End Sub

    ''' <summary> Bewirkt, dass nur Nummern in Textboxen akzeptiert werden. Außerdem wird die Input-Länge auf 5 limitiert </summary>
    Public Sub RegisterPlzInput(ParamArray plzTextBoxes As TextBox())
        For Each plztb In plzTextBoxes
            plztb.MaxLength = 5
            AddHandler plztb.KeyDown, AddressOf TextBox_numericKeyDown
            AddHandler plztb.Enter, AddressOf TextBox_Enter
            AddHandler plztb.MouseClick, AddressOf TextBox_MouseDoubleClick
            AddHandler plztb.Leave, AddressOf plztb_Leave
        Next
    End Sub

    Private Sub plztb_Leave(sender As Object, e As EventArgs)
        Dim tb = DirectCast(sender, TextBox)
        If Not tb.Text.Length = 5 Then
            msgExclamation("Wert muss 5-Stellig sein!")
            tb.Select() : tb.SelectAll()
        End If
    End Sub
#End Region 'PlzTextBox

#Region "JahrTextBox"

    Public Sub RegisterYearInput(ParamArray yearTextBoxes As TextBox())
        For Each ytb In yearTextBoxes
            ytb.MaxLength = 4
            AddHandler ytb.KeyDown, AddressOf TextBox_numericKeyDown
            AddHandler ytb.Enter, AddressOf TextBox_Enter
            AddHandler ytb.MouseClick, AddressOf TextBox_MouseDoubleClick
            AddHandler ytb.Leave, AddressOf ytb_Leave
        Next
    End Sub

    Private Sub ytb_Leave(sender As Object, e As EventArgs)
        Dim tb = DirectCast(sender, TextBox)
        If Not tb.Text.Length = 4 Then
            msgExclamation("Wert muss 4-Stellig sein!")
            tb.Select() : tb.SelectAll()
        End If
    End Sub

#End Region 'JahrTextBox

#Region "PZNTextBox"

    Public Sub RegisterPZNTextBoxInput(ParamArray pznTextBoxes As TextBox())
        For Each pznTb In pznTextBoxes
            AddHandler pznTb.KeyDown, AddressOf TextBox_numericKeyDown
            AddHandler pznTb.Enter, AddressOf TextBox_Enter
            AddHandler pznTb.MouseClick, AddressOf TextBox_MouseDoubleClick
            AddHandler pznTb.Leave, AddressOf TextBox_pznLeave
        Next
    End Sub

    Private Sub TextBox_pznLeave(sender As Object, e As EventArgs)
        Dim tb = DirectCast(sender, TextBox)
        Dim pzn = tb.Text
        Select Case pzn.Length
            Case 1 : tb.Text = $"0000000{pzn}"
            Case 2 : tb.Text = $"000000{pzn}"
            Case 3 : tb.Text = $"00000{pzn}"
            Case 4 : tb.Text = $"0000{pzn}"
            Case 5 : tb.Text = $"000{pzn}"
            Case 6 : tb.Text = $"00{pzn}"
            Case 7 : tb.Text = $"0{pzn}"
            Case 8 : tb.Text = pzn
            Case Else : tb.Text = "FEHLER!"
        End Select
    End Sub

#End Region

#Region "KennzeichenTextBox"

    Public Sub RegisterKennzeichenTextboxInput(ParamArray kennzeichenTextBoxes As TextBox())
        For Each ntb In kennzeichenTextBoxes
            AddHandler ntb.KeyDown, AddressOf TextBox_kennzeichenKeyDown
            AddHandler ntb.Enter, AddressOf TextBox_Enter
            AddHandler ntb.MouseClick, AddressOf TextBox_MouseDoubleClick
        Next
    End Sub

    Private Sub TextBox_kennzeichenKeyDown(sender As Object, e As KeyEventArgs)
        Dim tb = DirectCast(sender, TextBox)
        Select Case e.KeyCode
            Case Keys.D0 To Keys.D9, Keys.NumPad0 To Keys.NumPad9, Keys.Back 'nur Ziffern und Backspace zulassen
            Case Keys.A To Keys.Z
            Case Keys.Shift, Keys.Home, Keys.End, Keys.Delete, Keys.Left, Keys.Right, Keys.Space
            Case Else : e.SuppressKeyPress = True 'alle anderen Eingaben ablehnen
        End Select
    End Sub

#End Region 'KennzeichenTextBox

#Region "NumericTextBox"

    ''' <summary> Bewirkt, dass nur Zahlen in Textboxen akzeptiert werden. </summary>
    Public Sub RegisterNumericTextboxInput(ParamArray numericTextBoxes As TextBox())
        For Each ntb In numericTextBoxes
            AddHandler ntb.KeyDown, AddressOf TextBox_numericKeyDown
            AddHandler ntb.Enter, AddressOf TextBox_Enter
            AddHandler ntb.MouseClick, AddressOf TextBox_MouseDoubleClick
            AddHandler ntb.Leave, AddressOf TextBox_numericLeave
        Next
    End Sub

#End Region 'NumericTextBox

#Region "DateTextBox"

    ''' <summary>
    ''' bewirkt, dass nur ein Datum in die Textboxen eingegeben werden kann - akzeptiert auch verkürzte Datum-Syntax:
    ''' "d", "dd", "dMM", "ddMM", und Space (darüber wird direkt das aktuelle Datum in die Textbox eingefügt)
    ''' Überprüft beim Verlassen der TextBox NICHT, ob die Eingabe korrekt war.
    ''' </summary>
    ''' <param name="dateTextBoxes"></param>
    Public Sub RegisterDateInputNoCheck(ParamArray dateTextBoxes As TextBox())
        For Each dtb In dateTextBoxes
            AddHandler dtb.Enter, AddressOf TextBox_Enter
            AddHandler dtb.MouseClick, AddressOf TextBox_MouseDoubleClick
            AddHandler dtb.KeyDown, AddressOf dtb_KeyDown
            AddHandler dtb.Leave, AddressOf dtb_Leave_noCheck
        Next
    End Sub

    ''' <summary>
    ''' bewirkt, dass nur ein Datum in die Textboxen eingegeben werden kann - akzeptiert auch verkürzte Datum-Syntax:
    ''' "d", "dd", "dMM", "ddMM", und Space (darüber wird direkt das aktuelle Datum in die Textbox eingefügt)
    ''' Überprüft beim Verlassen der Textbox auf korrekte Eingabe.
    ''' </summary>
    ''' <param name="dateTextBoxes"></param>
    Public Sub RegisterDateInput(ParamArray dateTextBoxes As TextBox())
        For Each dtb In dateTextBoxes
            AddHandler dtb.Enter, AddressOf TextBox_Enter
            AddHandler dtb.MouseClick, AddressOf TextBox_MouseDoubleClick
            AddHandler dtb.KeyDown, AddressOf dtb_KeyDown
            AddHandler dtb.Leave, AddressOf dtb_Leave
        Next
    End Sub

    Private Sub dtb_KeyDown(sender As Object, e As KeyEventArgs)
        Dim tb = DirectCast(sender, TextBox)
        Select Case e.KeyCode
            Case Keys.Space : If Not tb.ReadOnly Then tb.Text = Date.Today.ToShortDateString
            Case Keys.Enter : tb.SelectNextControl(tb, True, True, True, True)
            Case Keys.D0 To Keys.D9, Keys.NumPad0 To Keys.NumPad9, Keys.OemPeriod, Keys.Back 'nur Ziffern "." und Backspace zulassen
            Case Keys.Shift, Keys.Home, Keys.End, Keys.Delete, Keys.Left, Keys.Right
            Case Else : e.SuppressKeyPress = True 'alle anderen Eingaben ablehnen
        End Select
    End Sub

    Private Sub dtb_Leave_noCheck(sender As Object, e As EventArgs)
        Dim dateResult As Date
        Dim tb = DirectCast(sender, TextBox)
        If DateTryParseSpecial(tb.Text, dateResult) Then
            tb.Text = dateResult.ToShortDateString
            tb.ForeColor = SystemColors.WindowText
        ElseIf tb.Text = "" Then
            tb.Text = "01.01.1900"
        End If
    End Sub

    Private Sub dtb_Leave(sender As Object, e As EventArgs)
        Dim dateResult As Date
        Dim tb = DirectCast(sender, TextBox)
        If DateTryParseSpecial(tb.Text, dateResult) Then
            tb.Text = dateResult.ToShortDateString
            tb.ForeColor = SystemColors.WindowText
        Else
            If tb.Text = "" Then
                tb.Text = "01.01.1900"
            Else
                tb.Text = "ungültiges Datum"
                tb.ForeColor = Color.Red
            End If
        End If
    End Sub

    Private Function DateTryParseSpecial(txt As String, ByRef result As Date) As Boolean
        If Date.TryParse(txt, result) Then Return True
        Dim chars = txt.ToCharArray
        Dim tag = "", monat = Date.Today.Month.ToString, jahr = Date.Today.Year.ToString
        Select Case txt.Length
            Case 1 : tag = {chars(0)}
            Case 2 : tag = {chars(0), chars(1)}
            Case 3 : tag = {chars(0), chars(1)} : monat = {chars(2)}
            Case 4, 6, 8
                tag = {chars(0), chars(1)} : monat = {chars(2), chars(3)}
                If txt.Length > 4 Then jahr = txt.Substring(4)
            Case Else : Return False
        End Select
        Return Date.TryParse($"{tag}.{monat}.{jahr}", result)
    End Function
#End Region 'DateTextBox

#Region "NumericComboBox"

    Public Sub RegisterNumericCmbInput(ParamArray cmbs As ComboBox())
        For Each cmb In cmbs
            AddHandler cmb.KeyDown, AddressOf NumericCmb_KeyDown
            AddHandler cmb.Leave, AddressOf NumericCmb_Leave
        Next
    End Sub

    Private Sub NumericCmb_KeyDown(sender As Object, e As KeyEventArgs)
        Dim cmb = DirectCast(sender, ComboBox)
        Select Case e.KeyCode
            Case Keys.D0 To Keys.D9, Keys.NumPad0 To Keys.NumPad9, Keys.Back 'nur Ziffern und Backspace zulassen
            Case Keys.Shift, Keys.Home, Keys.End, Keys.Delete, Keys.Left, Keys.Right
            Case Else : e.SuppressKeyPress = True 'alle anderen Eingaben ablehnen
        End Select
    End Sub

    Private Sub NumericCmb_Leave(sender As Object, e As EventArgs)
        Dim cmb = DirectCast(sender, ComboBox)
        If cmb.Text = "" Then cmb.Text = "0"
    End Sub

#End Region 'NumericComboBox

#Region "CheckArbeitszeit"

    Public Sub RegisterCheckArbeitszeit(ParamArray azTextBoxes As TextBox())
        For Each aztb In azTextBoxes
            AddHandler aztb.Leave, AddressOf aztb_Leave
        Next
    End Sub

    Private Sub aztb_Leave(sender As Object, e As EventArgs)
        Dim tb = DirectCast(sender, TextBox)
        Dim az = ConvertToDecimal(tb.Text)
        If az < 0 OrElse az > 24 Then
            msgExclamation("Die Arbeitszeit muss zwischen 0 und 24 liegen!")
            tb.Select() : tb.SelectAll()
        End If
    End Sub

#End Region 'CheckArbeitszeit

#Region "CheckKM"

    Public Sub RegisterCheckKM(ParamArray kmTextBoxes As TextBox())
        For Each kmtb In kmTextBoxes
            AddHandler kmtb.Leave, AddressOf kmtb_Leave
        Next
    End Sub

    Private Sub kmtb_Leave(sender As Object, e As EventArgs)
        Dim tb = DirectCast(sender, TextBox)
        Dim kmStand = ConvertToInteger(tb.Text)
        If kmStand < 0 OrElse kmStand > 999999 Then
            msgExclamation("Der KM-Stand muss zwischen 0 und 999999 liegen!")
            tb.Select() : tb.SelectAll()
        End If
    End Sub

#End Region 'CheckKM

#Region "SharedMethods"

    Private Sub TextBox_MouseDoubleClick(sender As Object, e As EventArgs)
        Dim tb = DirectCast(sender, TextBox)
        tb.SelectAll()
    End Sub

    Private Sub TextBox_Enter(sender As Object, e As EventArgs)
        Dim tb = DirectCast(sender, TextBox)
        tb.SelectAll()
    End Sub

    Private Sub PlzTextBox_KeyDown(sender As Object, e As KeyEventArgs)
        Dim tb = DirectCast(sender, TextBox)
        Select Case e.KeyCode
            Case Keys.Space : tb.Text = Date.Today.Year.ToString
            Case Keys.D0 To Keys.D9, Keys.NumPad0 To Keys.NumPad9, Keys.Back, Keys.X 'nur Ziffern, Backspace und "x" zulassen
            Case Keys.Shift, Keys.Home, Keys.End, Keys.Delete, Keys.Left, Keys.Right
            Case Else : e.SuppressKeyPress = True 'alle anderen Eingaben ablehnen
        End Select
    End Sub

    Private Sub TextBox_numericKeyDown(sender As Object, e As KeyEventArgs)
        Dim tb = DirectCast(sender, TextBox)
        Select Case e.KeyCode
            Case Keys.Space : tb.Text = Date.Today.Year.ToString
            Case Keys.D0 To Keys.D9, Keys.NumPad0 To Keys.NumPad9, Keys.Back 'nur Ziffern und Backspace zulassen
            Case Keys.Shift, Keys.Home, Keys.End, Keys.Delete, Keys.Left, Keys.Right
            Case Else : e.SuppressKeyPress = True 'alle anderen Eingaben ablehnen
        End Select
    End Sub

    Private Sub TextBox_numericLeave(sender As Object, e As EventArgs)
        Dim tb = DirectCast(sender, TextBox)
        If tb.Text = "" Then tb.Text = "0"
    End Sub

#End Region 'SharedMethods

#End Region 'TextBox

#Region "ToolStripMenuItem"

    <Extension>
    Public Sub CheckUncheckTS(check As ToolStripMenuItem, ParamArray uncheck As ToolStripMenuItem())
        check.Checked = True
        uncheck.ForEach(Sub(x) x.Checked = False)
    End Sub

    Public Sub SwitchTsCheckState(ParamArray TsItems As ToolStripMenuItem())
        For Each ts In TsItems
            If ts.Checked Then
                ts.Checked = False
            Else
                ts.Checked = True
            End If
        Next
    End Sub

    Public Sub CheckTs(ParamArray tsItems As ToolStripMenuItem())
        tsItems.ForEach(Sub(x) x.Checked = True)
    End Sub

    Public Sub UncheckTs(ParamArray tsItems As ToolStripMenuItem())
        tsItems.ForEach(Sub(x) x.Checked = False)
    End Sub

    Public Sub ShowTs(ParamArray tsItems As ToolStripMenuItem())
        tsItems.ForEach(Sub(x) x.Visible = True)
    End Sub

    Public Sub HideTs(ParamArray tsItems As ToolStripMenuItem())
        tsItems.ForEach(Sub(x) x.Visible = False)
    End Sub

    Public Sub DisableTs(ParamArray tsItems As ToolStripMenuItem())
        tsItems.ForEach(Sub(x) x.Enabled = False)
    End Sub

    Public Sub EnableTs(ParamArray tsItems As ToolStripMenuItem())
        tsItems.ForEach(Sub(x) x.Enabled = True)
    End Sub

#End Region 'ToolStripMenuItem

#Region "ToolStripStatusLabel"

    Public Sub ShowTss(ParamArray tssItems As ToolStripStatusLabel())
        tssItems.ForEach(Sub(x) x.Visible = True)
    End Sub

    Public Sub HideTss(ParamArray tssItems As ToolStripStatusLabel())
        tssItems.ForEach(Sub(x) x.Visible = False)
    End Sub

#End Region 'ToolStripStatusLabel

End Module
