﻿Imports System.Text, System.Text.RegularExpressions, System.Runtime.CompilerServices, System.Diagnostics
Imports System.Data.Common, MySql.Data

'TODO: MySQL Decimal Kommastellen einstellbar machen

Public Class DbGenerator

    Private _sCon As String
    Private _con As DbConnection
    Private _cmd As DbCommand
    Private _displayonly As Boolean
    Private _dts As DataSet
    Private _provider As Provider

    Public Sub New(providerData As ProviderData, dts As DataSet, displayonly As Boolean)
        Select Case providerData.Provider
            Case Provider.Access
                InitQuotes("` `")
                _con = New OleDb.OleDbConnection
                _cmd = New OleDb.OleDbCommand
            Case Provider.MsSQL
                InitQuotes("` `")
                _con = New MySqlClient.MySqlConnection
                _cmd = New MySqlClient.MySqlCommand
            Case Provider.MySQL
                InitQuotes("[ ]")
                _con = New SqlClient.SqlConnection
                _cmd = New SqlClient.SqlCommand
        End Select
        _sCon = providerData.sCon
        _cmd.Connection = _con
        _cmd.Connection.ConnectionString = _sCon
        _displayonly = displayonly
        _dts = dts
        _provider = providerData.Provider
    End Sub

    'TODO: noch für ganzes DataSet anbieten
    'Private Sub RunSQL(dts As DataSet, con As DbConnection, cmd As DbCommand, DisplayOnly As Boolean)
    '    If Not DisplayOnly Then
    '        Try
    '            ExecuteSql(GetBuildString(dts).ToArray, con, cmd)
    '        Catch ex As Exception
    '            Dim txt = Lf.Between(Lf, ex.Message, ex.InnerException.Message)
    '            MessageBox.Show(txt, ex.InnerException.GetType.Name, MessageBoxButtons.OK, MessageBoxIcon.Error)
    '        End Try
    '        msgInformation("Erstellen der Tabellenstruktur in der Datenbank abgeschlossen.")
    '    Else
    '        Using dlg As New dlgRtxt
    '            Dim rtb = dlg.RichTextBox1
    '            rtb.Lines = GetBuildString(dts).ToArray
    '            dlg.ShowDialog()
    '        End Using
    '    End If
    'End Sub

    Public Sub CreateTableAndColsAndIndexes(typedTable As DataTable, existingIndexnames As List(Of String))
        If Not _displayonly Then
            Try
                ExecuteSql(GetBuildString(typedTable, existingIndexnames), _con, _cmd, True)
            Catch ex As Exception
                Dim txt = Lf.Between(Lf, ex.Message, ex.InnerException.Message)
                MessageBox.Show(txt, ex.InnerException.GetType.Name, MessageBoxButtons.OK, MessageBoxIcon.Error)
            End Try
            msgInformation("Erstellen der Tabellenstruktur in der Datenbank abgeschlossen.")
        Else
            Using dlg As New dlgRtxt
                Dim rtb = dlg.RichTextBox1
                rtb.Lines = GetBuildString(typedTable, existingIndexnames).ToArray
                dlg.ShowDialog()
            End Using
        End If
    End Sub

    Private Sub ExecuteSql(sqlBuildUp As IEnumerable(Of String), con As DbConnection, cmd As DbCommand, singleTableOnly As Boolean)
        Using con
            con.Open()
            If Not singleTableOnly Then
                If msgQuestionWarning($"Es scheint, als soll das komplette DataSet in die Datenbank übermittelt werden.{Environment.NewLine}{Environment.NewLine}Soll die Datenbank vorher geleert werden? ACHTUNG! Vorgang kann nicht rückgängig gemacht werden!") = DialogResult.Yes Then
                    ClearDatabase(con, cmd)
                End If
            End If
            Cursor.Current = Cursors.WaitCursor
            ExecuteCommands(cmd, sqlBuildUp)
            Cursor.Current = Cursors.Default
        End Using
    End Sub

    Private Sub ExecuteCommands(cmd As DbCommand, sql As IEnumerable(Of String))
        For Each line In sql.Where(Function(ln) ln.Length > 3)
            Try
                cmd.CommandText = line
                cmd.ExecuteNonQuery()
                line = "Success".And(Lf, line)
            Catch ex As Exception
                msgEx(If(ex.InnerException.Null, ex, ex.InnerException), line)
            End Try
        Next
    End Sub

    Private Sub ClearDatabase(con As DbConnection, cmd As DbCommand)
        Using tb = con.GetSchema("Tables")
            Dim dropCommands = (From rw In tb.AsEnumerable Select "DROP TABLE ".And(rw("TABLE_NAME").ToString.Quote).And(" CASCADE")).ToArray
            Dim success = True
            While success
                success = False
                For i = dropCommands.Length - 1 To 0 Step -1
                    If dropCommands(i) = "" Then Continue For
                    Try
                        cmd.CommandText = dropCommands(i)
                        cmd.ExecuteNonQuery()
                        success = True
                        dropCommands(i) = ""
                    Catch ex As Exception
                        Debug.WriteLine("Alles gut, ist beim Löschen der Tables normal.....")
                    End Try
                Next
            End While
        End Using
    End Sub

    Private Function GetBuildString(typedTable As DataTable, existingIndexnames As List(Of String)) As List(Of String)
        Dim lines As New List(Of String)
        Dim lstStr As New List(Of String)
        Dim addRuleAction As Action(Of String, Rule) = Sub(rulename As String, rule As Rule)
                                                           If _provider = Provider.MsSQL Then 'MsSQL verlangt an der Stelle ausdrücklich ein "NO ACTION", bei anderen Providern kann das wegbleiben
                                                               If rule = Rule.None Then lstStr.Add($"ON {rulename} NO ACTION")
                                                               Return
                                                           Else
                                                               If rule = Rule.None Then Return
                                                           End If
                                                           lstStr.Add($"ON {rulename} {If(rule = Rule.Cascade, "CASCADE", If(rule = Rule.SetDefault, "SET DEFAULT", "SET NULL"))}")
                                                       End Sub
        Dim uniqueName As New Uniquifier
        Dim colHasNoExpression As Func(Of DataColumn, Boolean) = Function(col) col.Expression.HasNoValue
        If typedTable.PrimaryKey.Length = 0 Then
            msgExclamation($"Tabelle {typedTable.TableName} hat keinen PrimaryKey und wird übersprungen!")
            Return Nothing
        End If
        If typedTable.Columns.Count = 0 Then
            msgExclamation($"Tabelle {typedTable.TableName} hat keine Spalten und wird übersprungen!")
            Return Nothing
        End If
        lstStr.Clear()
        Select Case _provider
            Case Provider.Access : lstStr.Add($"CREATE TABLE {typedTable.TableName.Quote} (")
            Case Provider.MySQL : lstStr.Add($"CREATE TABLE {typedTable.TableName.Quote} (")
            Case Provider.MsSQL : lstStr.Add($"CREATE TABLE {typedTable.TableName.Quote} (")
        End Select
        For Each col In typedTable.ColumnsX.Where(colHasNoExpression)
            Dim colStr = GetColString(col)
            If colStr IsNot Nothing Then lstStr.Add(colStr)
        Next
        For Each uConstraint In typedTable.Constraints.OfType(Of UniqueConstraint)
            With uConstraint
                If .IsPrimaryKey Then
                    lstStr.Add($"CONSTRAINT {$"PK_{ .Table.TableName}".Quote} PRIMARY KEY")
                Else
                    lstStr.Add($"CONSTRAINT {uniqueName("U_", .Table.TableName).Quote} UNIQUE")
                End If
                lstStr.Add(.Columns.ColumngroupString)
            End With
            lstStr.Add(",")
        Next
        For Each fkConstraint In typedTable.Constraints.OfType(Of ForeignKeyConstraint)
            With fkConstraint
                lstStr.Add($"CONSTRAINT {uniqueName("FK_", .RelatedTable.TableName, "_", .Table.TableName).Quote}")
                lstStr.Add($"FOREIGN KEY { .Columns.ColumngroupString}")
                lstStr.Add($"REFERENCES { .RelatedTable.TableName.Quote}{ .RelatedColumns.ColumngroupString}")
                addRuleAction("UPDATE", .UpdateRule)
                addRuleAction("DELETE", .DeleteRule)
            End With
            lstStr.Add(",")
        Next
        If _provider = Provider.MySQL Then
            lstStr.Poke(");")
        Else
            lstStr.Poke(")")
        End If
        lines.Add(" ".Between(lstStr))
        lines.Add("")
        'Dim tbls = _dts.GetRankedTables.Where(Function(tbl) tbl.PrimaryKey.Length > 0).ToArray 'Tabellen ohne PrimKey-Column überspringen
        'Dim allUni = From tbl In tbls
        '             From uConstraint In tbl.Constraints.OfType(Of UniqueConstraint)
        '             Where uConstraint.Columns.All(colHasNoExpression)
        '             Select tbl = tbl.TableName, columns = $"{uConstraint.Columns.ColumngroupString}"
        'Dim allFK = From tbl In tbls
        '            From fkConstraint In tbl.Constraints.OfType(Of ForeignKeyConstraint)
        '            Where fkConstraint.RelatedColumns.All(colHasNoExpression)
        '            Select tbl = fkConstraint.RelatedTable.TableName, columns = $"{fkConstraint.RelatedColumns.ColumngroupString}"
        Dim allUni = From uConstraint In typedTable.Constraints.OfType(Of UniqueConstraint)
                     Where uConstraint.Columns.All(colHasNoExpression)
                     Select tbl = typedTable.TableName, columns = $"{uConstraint.Columns.ColumngroupString}"
        Dim allFK = From fkConstraint In typedTable.Constraints.OfType(Of ForeignKeyConstraint)
                    Where fkConstraint.RelatedColumns.All(colHasNoExpression)
                    Select tbl = fkConstraint.RelatedTable.TableName, columns = $"{fkConstraint.RelatedColumns.ColumngroupString}"
        For Each idx In allUni.Concat(allFK).Distinct
            If _provider = Provider.MySQL Then
                'lines.Add($"CREATE INDEX {uniqueName($"index{idx.tbl}").Quote} ON {idx.tbl.Quote}{idx.columns};")
                lines.Add($"CREATE INDEX {uniqueName.Uniquify($"index{idx.tbl}", existingIndexnames.AsEnumerable).Quote} ON {idx.tbl.Quote}{idx.columns};")
            Else
                'lines.Add($"CREATE INDEX {uniqueName($"index{idx.tbl}").Quote} ON {idx.tbl.Quote}{idx.columns}")
                lines.Add($"CREATE INDEX {uniqueName.Uniquify($"index{idx.tbl}", existingIndexnames.AsEnumerable).Quote} ON {idx.tbl.Quote}{idx.columns}")
            End If
        Next
        Return lines
    End Function

    Private Function ColOptionString(col As DataColumn, dbDependentName As String) As String
        Dim colStrBld As New StringBuilder
        Dim strAutoincrement = ""
        Select Case _provider
            Case Provider.Access : strAutoincrement = "COUNTER" '(1, 1)-was hat es damit auf sich??
            Case Provider.MySQL : strAutoincrement = $"{dbDependentName} AUTO_INCREMENT"
            Case Provider.MsSQL : strAutoincrement = "INT IDENTITY"
        End Select
        Dim strNotNull = " NOT NULL"
        Dim strNull = " NULL"
        Dim strUnique = " UNIQUE"
        If col.AutoIncrement Then colStrBld.Append(strAutoincrement) : dbDependentName = ""
        If Not col.AutoIncrement AndAlso col.Unique Then colStrBld.Append(strUnique) 'Unique nur hinzufügen, wenn Col nicht sowieso ID(mit Autoinkrement) ist
        If Not col.AllowDBNull Then colStrBld.Append(strNotNull)
        Return $"{col.ColumnName.Quote} {dbDependentName}{colStrBld},"
    End Function

End Class

Public Class ProviderData

    Public Property Provider As Provider
    Public Property ServerOrFile As String
    Public Property DbName As String
    Public Property Username As String
    Public Property Pass As String
    Public Property sCon As String

    Public Sub New(SelectedProvider As Provider, ServerOrDbfile As String, DbN As String, UsrN As String, PW As String)
        ServerOrFile = ServerOrDbfile
        DbName = DbN
        Username = UsrN
        Pass = PW
        Provider = SelectedProvider
        Select Case SelectedProvider
            Case Provider.Access
                sCon = $"Provider=Microsoft.Jet.OLEDB.4.0;Data Source={ServerOrFile}{If(Pass = "", "", $";Jet OLEDB:Database Password={Pass}")}"
            Case Provider.MsSQL
                sCon = $"Server={ServerOrFile}; UID={Username}; password={Pass}; Initial Catalog={DbName};"
            Case Provider.MySQL
                sCon = $"Server={ServerOrFile}; Database={DbName}; User ID={Username}; Password={Pass};"
        End Select
    End Sub

End Class

Public Enum Provider
    Access = 1
    MsSQL = 2
    MySQL = 3
End Enum

Module CreateTableHelpers

#Region "IList"

    <Extension(), DebuggerStepThrough()>
    Public Sub Add(Of T)(ByVal subj As ICollection(Of T), ParamArray itms() As T)
        For Each itm In itms : subj.Add(itm) : Next
    End Sub

    ''' <summary>Bsp: setze vorletztes Element: list.Poke(-1, "huhu")</summary>
    <Extension(), DebuggerStepThrough()>
    Public Sub Poke(Of T)(ByVal subj As IList(Of T), ByVal value As T, Optional ByVal index As Integer = 0)
        subj(index + subj.Count - 1) = value
    End Sub

#End Region 'IList

    Dim _InnerQuotes As String
    Dim _OuterQuotes As String()
    Dim _Parenthesis As String() = {"( ", " )"}

    Sub InitQuotes(quotes As String)
        _OuterQuotes = quotes.Split
        _InnerQuotes = $"{_OuterQuotes(1)},{_OuterQuotes(0)}"
    End Sub

    <Extension()>
    Function ColumngroupString(cols As DataColumn()) As String
        Return _InnerQuotes.Between(cols).Between(_OuterQuotes).InParenthesis
    End Function

    <Extension()>
    Public Function InParenthesis(ByVal Subj As String) As String
        Return Subj.Between(_Parenthesis)
    End Function

    <Extension()>
    Public Function Quote(ByVal Subj As String) As String
        Return Subj.Between(_OuterQuotes)
    End Function

End Module

Public Class Uniquifier : Inherits Dictionary(Of String, Integer)

    Private Shared _rgxLastDigits As New Regex(" \ d + $", RegexOptions.RightToLeft)

    ''' <summary> appends a number at baseword, to ensure, that its unique in the previousUniques-List  </summary> 
    Public Function Uniquify(baseWord As String, previousUniques As IEnumerable(Of String)) As String
        Dim mt = _rgxLastDigits.Match(baseWord)
        If mt.Success Then baseWord = baseWord.Substring(0, mt.Index)
        Dim i = If(mt.Success, Integer.Parse(mt.Value), 0)
        For Each wrd In previousUniques
            mt = _rgxLastDigits.Match(wrd)
            If mt.Success Then wrd = wrd.Substring(0, mt.Index)
            If wrd = baseWord Then i = Math.Max(i, Integer.Parse(mt.Value))
        Next
        Return baseWord & i + 1
    End Function

    ''' <summary> appends a number at baseword, to ensure, that its unique in this instance of Uniquifier </summary> 
    Default Public Overloads ReadOnly Property Item(baseword As String) As String
        Get
            baseword = _rgxLastDigits.Replace(baseword, "")
            Dim i = 0
            If TryGetValue(baseword, i) Then i += 1
            MyBase.Item(baseword) = i
            Return baseword & i
        End Get
    End Property

    ''' <summary> appends a number at baseword, to ensure, that its unique in this instance of Uniquifier </summary> 
    Default Public Overloads ReadOnly Property Item(baseword As String, ParamArray basewordSegments() As String) As String
        Get
            Return Item(baseword & String.Concat(basewordSegments))
        End Get
    End Property

End Class

Public Module DbGeneratorMod 'TODO: alter DB-Generator - kann weg, wenn der neue funktioniert

    Private _provider As Provider
    Private _sCon As String
    Private _displayOnly As Boolean

    <Extension>
    Public Sub DbGenerator(dts As DataSet, providerData As ProviderData)
        Using dlg As New dlgDbGenerator(providerData)
            If dlg.ShowDialog <> DialogResult.OK Then Return
            _displayOnly = dlg.rbtnCmdOnly.Checked
            Select Case dlg._SelectedProvider
                Case "Access"
                    _provider = Provider.Access
                    InitQuotes("` `")
                    _sCon = $"Provider=Microsoft.Jet.OLEDB.4.0;Data Source={dlg.tbServerOrFile.Text}{If(dlg.tbPass.Text = "", "", $";Jet OLEDB:Database Password={dlg.tbPass.Text}")}"
                    Dim con As New OleDb.OleDbConnection
                    Dim cmd As New OleDb.OleDbCommand
                    cmd.Connection = con
                    cmd.Connection.ConnectionString = _sCon
                    RunSQL(dts, con, cmd, _displayOnly)
                Case "MySQL"
                    _provider = Provider.MySQL
                    InitQuotes("` `")
                    _sCon = $"Server={dlg.tbServerOrFile.Text}; UID={dlg.tbUsername.Text}; password={dlg.tbPass.Text}; Initial Catalog={dlg.tbDbName.Text};"
                    Dim con As New MySqlClient.MySqlConnection
                    Dim cmd As New MySqlClient.MySqlCommand
                    cmd.Connection = con
                    cmd.Connection.ConnectionString = _sCon
                    RunSQL(dts, con, cmd, _displayOnly)
                Case "MsSQL"
                    _provider = Provider.MsSQL
                    InitQuotes("[ ]")
                    _sCon = $"Server={dlg.tbServerOrFile.Text}; Database={dlg.tbDbName.Text}; User ID={dlg.tbUsername.Text}; Password={dlg.tbPass.Text};"
                    Dim con As New SqlClient.SqlConnection
                    Dim cmd As New SqlClient.SqlCommand
                    RunSQL(dts, con, cmd, _displayOnly)
            End Select
        End Using
    End Sub

    Private Sub RunSQL(dts As DataSet, con As DbConnection, cmd As DbCommand, DisplayOnly As Boolean)
        If Not DisplayOnly Then
            Try
                ExecuteSql(GetBuildString(dts).ToArray, con, cmd)
            Catch ex As Exception
                Dim txt = Lf.Between(Lf, ex.Message, ex.InnerException.Message)
                MessageBox.Show(txt, ex.InnerException.GetType.Name, MessageBoxButtons.OK, MessageBoxIcon.Error)
            End Try
            msgInformation("Erstellen der Tabellenstruktur in der Datenbank abgeschlossen.")
        Else
            Using dlg As New dlgRtxt
                Dim rtb = dlg.RichTextBox1
                rtb.Lines = GetBuildString(dts).ToArray
                dlg.ShowDialog()
            End Using
        End If
    End Sub

    Private Sub ExecuteSql(sqlBuildUp As IEnumerable(Of String), con As DbConnection, cmd As DbCommand)
        Using con
            con.Open()
            If msgQuestionInfo("Soll die Datenbank vorher geleert werden?") = DialogResult.Yes Then
                ClearDB(con, cmd)
            End If
            Cursor.Current = Cursors.WaitCursor
            ExecuteCommands(cmd, sqlBuildUp)
            Cursor.Current = Cursors.Default
        End Using
    End Sub

    Private Sub ExecuteCommands(cmd As DbCommand, sql As IEnumerable(Of String))
        For Each line In sql.Where(Function(ln) ln.Length > 3)
            Try
                cmd.CommandText = line
                cmd.ExecuteNonQuery()
                line = "Success".And(Lf, line)
            Catch ex As Exception
                msgEx(If(ex.InnerException.Null, ex, ex.InnerException), line)
            End Try
        Next
    End Sub

    Private Sub ClearDB(con As DbConnection, cmd As DbCommand)
        Using tb = con.GetSchema("Tables")
            Dim dropCommands = (From rw In tb.AsEnumerable Select "DROP TABLE ".And(rw("TABLE_NAME").ToString.Quote).And(" CASCADE")).ToArray
            Dim success = True
            While success
                success = False
                For i = dropCommands.Length - 1 To 0 Step -1
                    If dropCommands(i) = "" Then Continue For
                    Try
                        cmd.CommandText = dropCommands(i)
                        cmd.ExecuteNonQuery()
                        success = True
                        dropCommands(i) = ""
                    Catch ex As Exception
                        Debug.WriteLine("Alles gut, ist beim Löschen der Tables normal.....")
                    End Try
                Next
            End While
        End Using
    End Sub

    Private Function GetBuildString(dts As DataSet) As List(Of String)
        Dim lines As New List(Of String)
        Dim lstStr As New List(Of String)
        Dim addRuleAction As Action(Of String, Rule) = Sub(rulename As String, rule As Rule)
                                                           If _provider = Provider.MsSQL Then 'MsSQL verlangt an der Stelle ausdrücklich ein "NO ACTION", bei anderen Providern kann das wegbleiben
                                                               If rule = Rule.None Then lstStr.Add($"ON {rulename} NO ACTION")
                                                               Return
                                                           Else
                                                               If rule = Rule.None Then Return
                                                           End If
                                                           lstStr.Add($"ON {rulename} {If(rule = Rule.Cascade, "CASCADE", If(rule = Rule.SetDefault, "SET DEFAULT", "SET NULL"))}")
                                                       End Sub
        Dim uniqueName As New Uniquifier
        Dim colHasNoExpression As Func(Of DataColumn, Boolean) = Function(col) col.Expression.HasNoValue
        Dim tbls = dts.GetRankedTables.Where(Function(tbl) tbl.PrimaryKey.Length > 0).ToArray 'Tabellen ohne PrimKey-Column überspringen
        For Each tbl As DataTable In tbls
            If tbl.Columns.Count = 0 Then Continue For
            lstStr.Clear()
            Select Case _provider
                Case Provider.Access : lstStr.Add($"CREATE TABLE {tbl.TableName.Quote} (") 'IF NOT EXISTS testen
                Case Provider.MySQL : lstStr.Add($"CREATE TABLE {tbl.TableName.Quote} (") 'IF NOT EXISTS testen
                Case Provider.MsSQL : lstStr.Add($"CREATE TABLE {tbl.TableName.Quote} (")
            End Select
            For Each col In tbl.Columns.Cast(Of DataColumn).Where(colHasNoExpression)
                Dim colStr = GetColString(col)
                If colStr IsNot Nothing Then lstStr.Add(colStr)
            Next
            For Each uConstraint In tbl.Constraints.OfType(Of UniqueConstraint)
                With uConstraint
                    If .IsPrimaryKey Then
                        lstStr.Add($"CONSTRAINT {$"PK_{ .Table.TableName}".Quote} PRIMARY KEY")
                    Else
                        lstStr.Add($"CONSTRAINT {uniqueName("U_", .Table.TableName).Quote} UNIQUE")
                    End If
                    lstStr.Add(.Columns.ColumngroupString)
                End With
                lstStr.Add(",")
            Next
            For Each fkConstraint In tbl.Constraints.OfType(Of ForeignKeyConstraint)
                With fkConstraint
                    lstStr.Add($"CONSTRAINT {uniqueName("FK_", .RelatedTable.TableName, "_", .Table.TableName).Quote}")
                    lstStr.Add($"FOREIGN KEY { .Columns.ColumngroupString}")
                    lstStr.Add($"REFERENCES { .RelatedTable.TableName.Quote}{ .RelatedColumns.ColumngroupString}")
                    addRuleAction("UPDATE", .UpdateRule)
                    addRuleAction("DELETE", .DeleteRule)
                End With
                lstStr.Add(",")
            Next
            If _provider = Provider.MySQL Then
                lstStr.Poke(");")
            Else
                lstStr.Poke(")")
            End If
            lines.Add(" ".Between(lstStr))
            lines.Add("")
        Next
        Dim allUni = From tbl In tbls
                     From uConstraint In tbl.Constraints.OfType(Of UniqueConstraint)
                     Where uConstraint.Columns.All(colHasNoExpression)
                     Select tbl = tbl.TableName, columns = $"{uConstraint.Columns.ColumngroupString}"
        Dim allFK = From tbl In tbls
                    From fkConstraint In tbl.Constraints.OfType(Of ForeignKeyConstraint)
                    Where fkConstraint.RelatedColumns.All(colHasNoExpression)
                    Select tbl = fkConstraint.RelatedTable.TableName, columns = $"{fkConstraint.RelatedColumns.ColumngroupString}"
        For Each idx In allUni.Concat(allFK).Distinct
            If _provider = Provider.MySQL Then
                lines.Add($"CREATE INDEX {uniqueName($"index{idx.tbl}").Quote} ON {idx.tbl.Quote}{idx.columns};")
            Else
                lines.Add($"CREATE INDEX {uniqueName($"index{idx.tbl}").Quote} ON {idx.tbl.Quote}{idx.columns}")
            End If
        Next
        Return lines
    End Function

    Private Function ColOptionString(col As DataColumn, dbDependentName As String) As String
        Dim colStrBld As New StringBuilder
        Dim strAutoincrement = ""
        Select Case _provider
            Case Provider.Access : strAutoincrement = "COUNTER" '(1, 1)-was hat es damit auf sich??
            Case Provider.MySQL : strAutoincrement = $"{dbDependentName} AUTO_INCREMENT"
            Case Provider.MsSQL : strAutoincrement = "INT IDENTITY"
        End Select
        Dim strNotNull = " NOT NULL"
        Dim strNull = " NULL"
        Dim strUnique = " UNIQUE"
        If col.AutoIncrement Then colStrBld.Append(strAutoincrement) : dbDependentName = ""
        If Not col.AutoIncrement AndAlso col.Unique Then colStrBld.Append(strUnique) 'Unique nur hinzufügen, wenn Col nicht sowieso ID(mit Autoinkrement) ist
        If Not col.AllowDBNull Then colStrBld.Append(strNotNull)
        Return $"{col.ColumnName.Quote} {dbDependentName}{colStrBld},"
    End Function

    Public Function GetColString(col As DataColumn) As String
        Dim dbDependentName As String = ""
        Select Case col.DataType
            Case GetType(Boolean)
                Select Case _provider
                    Case Provider.MySQL : dbDependentName = "TINYINT"
                    Case Provider.MsSQL : dbDependentName = "bit"
                    Case Provider.Access : dbDependentName = "Bit"
                End Select
            Case GetType(Byte)
                Select Case _provider
                    Case Provider.MySQL : dbDependentName = "TINYINT UNSIGNED"
                    Case Provider.MsSQL : dbDependentName = Nothing
                    Case Provider.Access : dbDependentName = "Byte"
                End Select
            Case GetType(Byte())
                Select Case _provider
                    Case Provider.MySQL : dbDependentName = "varbinary"
                    Case Provider.MsSQL : dbDependentName = "image"
                    Case Provider.Access : dbDependentName = "LongBinary"
                End Select
            Case GetType(Char)
                Select Case _provider
                    Case Provider.MySQL : dbDependentName = "NCHAR"
                    Case Provider.MsSQL : dbDependentName = "nchar"
                    Case Provider.Access : dbDependentName = "VarChar"
                End Select
            Case GetType(Date)
                Select Case _provider
                    Case Provider.MySQL : dbDependentName = "DATETIME"
                    Case Provider.MsSQL : dbDependentName = "datetime"
                    Case Provider.Access : dbDependentName = "DateTime"
                End Select
            Case GetType(DateTimeOffset)
                Select Case _provider
                    Case Provider.MySQL : dbDependentName = Nothing
                    Case Provider.MsSQL : dbDependentName = Nothing
                    Case Provider.Access : dbDependentName = Nothing
                End Select
            Case GetType(Decimal)
                Select Case _provider
                    Case Provider.MySQL : dbDependentName = "DECIMAL(10,2)"
                    Case Provider.MsSQL : dbDependentName = "decimal(18,2)"
                    Case Provider.Access : dbDependentName = "Double"
                End Select
            Case GetType(Double)
                Select Case _provider
                    Case Provider.MySQL : dbDependentName = "DOUBLE"
                    Case Provider.MsSQL : dbDependentName = "float"
                    Case Provider.Access : dbDependentName = "Double"
                End Select
            Case GetType(Guid)
                Select Case _provider
                    Case Provider.MySQL : dbDependentName = Nothing
                    Case Provider.MsSQL : dbDependentName = "uniqueidentifier"
                    Case Provider.Access : dbDependentName = "GUID"
                End Select
            Case GetType(Short)
                Select Case _provider
                    Case Provider.MySQL : dbDependentName = "SMALLINT"
                    Case Provider.MsSQL : dbDependentName = "smallint"
                    Case Provider.Access : dbDependentName = "Short"
                End Select
            Case GetType(Integer)
                Select Case _provider
                    Case Provider.MySQL : dbDependentName = "INT"
                    Case Provider.MsSQL : dbDependentName = "int"
                    Case Provider.Access : dbDependentName = "Long"
                End Select
            Case GetType(Long)
                Select Case _provider
                    Case Provider.MySQL : dbDependentName = "BIGINT"
                    Case Provider.MsSQL : dbDependentName = "bigint"
                    Case Provider.Access : dbDependentName = Nothing
                End Select
            Case GetType(Object)
                Select Case _provider
                    Case Provider.MySQL : dbDependentName = Nothing
                    Case Provider.MsSQL : dbDependentName = Nothing
                    Case Provider.Access : dbDependentName = Nothing
                End Select
            Case GetType(SByte)
                Select Case _provider
                    Case Provider.MySQL : dbDependentName = "TINYINT"
                    Case Provider.MsSQL : dbDependentName = "tinyint"
                    Case Provider.Access : dbDependentName = Nothing
                End Select
            Case GetType(Single)
                Select Case _provider
                    Case Provider.MySQL : dbDependentName = "FLOAT"
                    Case Provider.MsSQL : dbDependentName = "real"
                    Case Provider.Access : dbDependentName = "Single"
                End Select
            Case GetType(String)
                Select Case _provider
                    Case Provider.MySQL : dbDependentName = If(col.MaxLength = -1, "VARCHAR(255)", "TEXT") 'Mediumtext: 16,777,215 Chars, Longtext: 4,294,967,295 Chars
                    Case Provider.MsSQL : dbDependentName = If(col.MaxLength = -1, "text", If(col.MaxLength <= 8000, $"varchar({col.MaxLength})", "text"))
                    Case Provider.Access : dbDependentName = If(col.MaxLength = -1, "LongText", If(col.MaxLength > 64000, "LongText", "VarChar"))
                End Select
            Case GetType(TimeSpan)
                Select Case _provider
                    Case Provider.MySQL : dbDependentName = "TIME"
                    Case Provider.MsSQL : dbDependentName = Nothing
                    Case Provider.Access : dbDependentName = Nothing
                End Select
            Case GetType(UShort)
                Select Case _provider
                    Case Provider.MySQL : dbDependentName = "SMALLINT UNSIGNED"
                    Case Provider.MsSQL : dbDependentName = Nothing
                    Case Provider.Access : dbDependentName = Nothing
                End Select
            Case GetType(UInteger)
                Select Case _provider
                    Case Provider.MySQL : dbDependentName = "MEDIUMINT UNSIGNED"
                    Case Provider.MsSQL : dbDependentName = Nothing
                    Case Provider.Access : dbDependentName = Nothing
                End Select
            Case GetType(ULong)
                Select Case _provider
                    Case Provider.MySQL : dbDependentName = "BIGINT UNSIGNED"
                    Case Provider.MsSQL : dbDependentName = Nothing
                    Case Provider.Access : dbDependentName = Nothing
                End Select
        End Select
        If dbDependentName Is Nothing Then
            msgInformation($"Der Datentyp {col.DataType} für Spalte {col.ColumnName} ist nicht implementiert, wird übersprungen!")
            Return Nothing
        End If
        Return ColOptionString(col, dbDependentName)
    End Function

End Module