﻿Imports System.Runtime.CompilerServices
Public Module MiscT

    ''' <summary>erzeugt aus einem hierarchischen InhaltsVerzeichnis-Eintrag (TableOfContent - TOC) einen sich lexikografisch korrekt einordnenden SortKey.
    '''  Beispiel: "2.11.3 Blabla" konvertiert zu " 2 B 3" (verkettete 2-stellig-rechtsbündige Hex-Darstellung der Segmente)
    '''  MaxValue eines tocEntry-Segments ist folglich die grösste 2-stellige hex-darstellung: FF, in decimal: 255</summary>
    Public Function GetTreeViewSortKey(tocEntry As String) As String
        Dim spaceIndex = tocEntry.IndexOf(" "c)
        Select Case spaceIndex
            Case 0 : Throw New ArgumentException("mustn't start with ' '", NameOf(tocEntry))
            Case > 0 : tocEntry = tocEntry.Substring(0, spaceIndex)
        End Select
        Dim numbs = tocEntry.Split("."c).Select(AddressOf Integer.Parse).ToArray
        If numbs.Any(Function(n) n > Byte.MaxValue) Then Throw New ArgumentException("segments maximum is '255'", NameOf(tocEntry))
        Return String.Concat(numbs.Select(Function(n) $"{n,2:X}")) ' 2-stellig rechtsbündige Hex-Darstellung
    End Function

    Public Function GetTocSortKey(tocEntry As String) As String
        Dim splits = tocEntry.Split(". ".ToCharArray, StringSplitOptions.RemoveEmptyEntries)
        Return String.Concat(splits.Select(Function(x) $"{x,15}")) ' rechtsbündig in Spalte mit Breite 15
    End Function

    ''' <summary>Gibt zurück, ob ein StringArray nur leere Strings ("") enthält, oder nicht</summary>
    <Extension>
    Public Function IsEmpty(StringArray As String()) As Boolean
        Dim empty = False
        For Each s In StringArray
            If s = "" Then
                empty = True
            Else
                empty = False
            End If
        Next
        Return empty
    End Function

    ''' <summary>Gibt den längsten String eines StringArrays zurück</summary>
    <Extension>
    Public Function GetLongestString(SomeStrings As String()) As String
        Dim longest = ""
        For Each str In SomeStrings
            If str.Length > longest.Length Then longest = str
        Next
        Return longest
    End Function

End Module
